/*
 * Copyright (C) 2003 Wolfgang Moser <womo_(at)_d81_(dot)_de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
  Check Commodore 1541 style DOS ROMs for correct ROM checksums

    - If the checksum seems to be incorrect, make proposions, how
      to fix it

    - Analyse the file size and automatically decide, if the ROM
      image would be a plain binary ROM file or a typical CBM style
      EPROM burner PRG file with a load address attached

    - It is always assumed, that the ROM image is <= 64kB
      (not including the load address) and that it is consecutively
      located in the upper most memory area (from $0000 to $ffff).

    - If the ROM image is 64kB big (excluding the load address), the
      checksum of the very first 8kB block can never become $00,
      except, if all bytes of this area are $00. This is a design
      flaw of the CBM checksum routine, which adds the carry bit of
      an 8-bit addition to the same sum again. Therefore e.g.
      $ff + $01 results in $01 except of $00. You won't find a pair
      of operands, that will result in a sum of $00, except if you
      choose $00 and $00.

    - The ROM size must always be a multiple of 8kB (not taking the
      load address into account)
*/

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

using namespace std;	// Needed for GCC-3 and its standard headers

  // main EPROMfile
int main(int argc, char** argv){
  bool                    overall;
  unsigned char           bbuf[0x2000];
  unsigned int            i, j, cksum;
  unsigned long int       bflen;

  cout << "This tool comes with ABSOLUTELY NO WARRANTY; "
          "read COPYING for details.\n" << endl;

  if(argc!=2){
    cerr << "1541ROM-Check CBM-DOS-ROM" << endl;
    return 1;
    }

  ifstream bfin(argv[1],ios::in);
  if(!bfin){
    cerr << "Couldn't open ROM input file \""
         << argv[1] << "\"." << endl;
    return 2;
    }

  bfin.seekg(0,ios::end);
  bflen=bfin.tellg();

  switch(bflen & 0x1fff){
    default:	// ROM file is no multiple of 8kB
      cerr << "The ROM input file is not a multiple of 8kB" << endl;
      return 3;
    case 2:	// ROM file with load address
      cout << '"' << argv[1] << "\" is a CBM style ROM file with"
                                " a load address," << endl;
      break;
    case 0:	// plain binary ROM file without load address
      cout << '"' << argv[1] << "\" is a plain binary ROM file," << endl;
    }

  cout << setfill('0') << hex << "it fills the memory area from 0x"
       << setw(2) << (0x100-(bflen>>8)) << "00 to 0xffff\n" << endl;

    // seek to the first 8kB block to read
  bfin.seekg(bflen & 0x1fff,ios::beg);

  overall=true;

    // i - checksum to reach, e.g.:  0xC0, 0xE0
  for(i=0x100-(bflen>>8); i<0x100; i+=0x20){
      // load a 8kB block into the buffer

    if(!bfin.read(reinterpret_cast<char *>(bbuf), 0x2000) || bfin.gcount()!=0x2000){
      cerr << "Unable to read the next 8kB block from the file \""
           << argv[1] << "\"." << endl;
      }
    cout << "Checking the 8kB ROM area from 0x"
         << setw(2) << i << "00 to 0x" << setw(2) << (i+0x1f)
         << "ff\nfor the correct checksum value of 0x" << i
         << endl;

    if(i==0x00){
      cout << "WARNING: A checksum of 0x00 cannot be reached,\nexcept if"
              "all bytes of this area are 0x00 itsef" << endl;
      }

      // calculate the 16-bit sum
    cksum=0;
#if 1
    for(j=0; j<0x2000; j++) cksum+=bbuf[j];
      // reduce to an 8-bit sum
    while(cksum>0xff) cksum=(cksum&0xff) + (cksum>>8);
#else
    for(j=0; j<0x2000; j++){
      cksum+=bbuf[j];
        // reduce to 8-bit (just like the original CBM routine)
      if(cksum>0xff){
	cksum-=0xff;	// decrease by 0x100, but add the carry
	}
      }
#endif
    // The reduction to 8 bits means, that the checksum can never
    // become 0 itself, except if all bytes added are 0!!!


      // check for correctness
    if(cksum==i){
      cout << "\n The checksum is OK.\n" << endl;
      }
    else{
      overall=false;
      cout << "\n The checksum is WRONG! The current value is: 0x"
           << setw(2) << cksum << endl;

      j=(i-cksum);
      j=(j&0xff)-(j>0xff);

      cout << "  To fix it, wou will have to increase one byte of this\n"
              "  ROM block by 0x" << setw(2) << j <<
              "  (or 0x" << (j+1) << ", if the 8-bit addition\n"
              "  would set the carry bit).\n" << endl;
      switch(i){
        case 0xE0:
        case 0xC0:
          cout << "  If this ROM should be a (slightly modified) original\n"
                  "  Commodore disk drive ROM, you would have to change the\n"
                  "  byte at address location ";
        }
      switch(i){
        case 0xE0:
          j+=bbuf[0x1EE6];
          if(j>0xff) j-=0xff;
          cout << "0xFEE6 from 0x" << setw(2) << static_cast<unsigned int>(bbuf[0x1EE6])
               << " to 0x" << setw(2) << j << ".\n\n" << endl;
          break;
        case 0xC0:
          j+=bbuf[0x0001];
          if(j>0xff) j-=0xff;
          cout << "0xC001 from 0x" << setw(2) << static_cast<unsigned int>(bbuf[0x0001])
               << " to 0x" << setw(2) << j << ".\n\n" << endl;
          break;
        }
      }
    }

  cout << dec << setfill(' ') << "\nThe overall ROM checksum is:   ";
  if(overall) cout << "correct." << endl;
  else cout << "NOT   correct" << endl;
  return 0;
  }
